from multiprocessing import Process, JoinableQueue
from Queue import Empty
import time
import re

#essence_line_pattern = re.compile(r"""^( )*(?P<name>[a-zA-Z'\- \(\)]+\b)( )*\t( )*[^\t]*\t( )*[^\t]*\t( )*[^\t]*\t( )*(?P<Aromatic>-?[0-3])( )*\t( )*(?P<Astringent>-?[0-3])( )*\t( )*(?P<Bitter>-?[0-3])( )*\t( )*(?P<Salty>-?[0-3])( )*\t( )*(?P<Sour>-?[0-3])( )*\t( )*(?P<Spicy>-?[0-3])( )*\t( )*(?P<Sweet>-?[0-3])( )*\t( )*(?P<Toxic>-?[0-3])( )*\t( )*(?P<cost>\d+)""")
essence_line_pattern = re.compile(r"""^( )*(?P<name>[a-zA-Z'\- \(\)]*[a-zA-Z'\- \(\)])( )*\t( )*(?P<Aromatic>-?[0-3])( )*\t( )*(?P<Astringent>-?[0-3])( )*\t( )*(?P<Bitter>-?[0-3])( )*\t( )*(?P<Salty>-?[0-3])( )*\t( )*(?P<Sour>-?[0-3])( )*\t( )*(?P<Spicy>-?[0-3])( )*\t( )*(?P<Sweet>-?[0-3])( )*\t( )*(?P<Toxic>-?[0-3])( )*\t( )*(?P<cost>\d+)""")

AttrNames = ["Aromatic", "Astringent", "Bitter", "Salty", "Sour", "Spicy", "Sweet", "Toxic"]
AttrShortNames = ["Ar", "As", "Bi", "Sa", "So", "Sp", "Sw", "To"]
NumAttr=8
NumEVals=7
NumE=5

def permute_next(values):
    """
    Alter the list of values in-place to produce to the next permutation
    in lexicographical order.
    
    'values' must support slicing and ::reverse().
    """
    last = len(values) - 1
    a = last
    while a > 0:
        b = a
        a -= 1
        if values[a] < values[b]:
            c = last
            while values[a] >= values[c]: # >= allows duplicates
                c -= 1
            values[a], values[c] = values[c], values[a] # Swap.
            values[b:] = values[:b-1:-1] # Reverse.
            return
    values.reverse()

class Essence:
    def __init__ (self, m, essences2, essences3):
        self.name = m.group('name')
        self.cost = int(m.group('cost'))
        #self.cost = 0
        
        self.values = [0]*8

        self.values[0] = int(m.group(AttrNames[0])) + 3
        self.values[1] = int(m.group(AttrNames[1])) + 3
        self.values[2] = int(m.group(AttrNames[2])) + 3
        self.values[3] = int(m.group(AttrNames[3])) + 3
        self.values[4] = int(m.group(AttrNames[4])) + 3
        self.values[5] = int(m.group(AttrNames[5])) + 3
        self.values[6] = int(m.group(AttrNames[6])) + 3
        self.values[7] = int(m.group(AttrNames[7])) + 3

        for attr1 in range(NumAttr-1):
            for attr2 in range(NumAttr-1-attr1):
                essences2[attr1][attr2][self.values[attr1]][self.values[attr2+attr1+1]].append(self)
        for attr1 in range(NumAttr-2):
            for attr2 in range(NumAttr-2-attr1):
                for attr3 in range(NumAttr-2-attr2-attr1):
                    essences3[attr1][attr2][attr3][self.values[attr1]][self.values[attr2+attr1+1]][self.values[attr3+attr2+attr1+2]].append(self)
    
    def __str__ (self):
        return self.name

    def __cmp__ (self, other):
        assert isinstance(other, Essence)
        result = cmp(self.cost, other.cost)
        if result == 0:
            result = cmp(self.name, other.name)
        return result

# Build structure of the Essence tables
def buildEssenceTables(essences2, essences3):
    for attr1 in range(NumAttr-2):
        essences3.append([])
        for attr2 in range(NumAttr-2-attr1):
            essences3[attr1].append([])
            for attr3 in range(NumAttr-2-attr2-attr1):
                essences3[attr1][attr2].append([])
                for val1 in range(NumEVals):
                    essences3[attr1][attr2][attr3].append([])
                    for val2 in range(NumEVals):
                        essences3[attr1][attr2][attr3][val1].append([])
                        for val3 in range(NumEVals):
                            essences3[attr1][attr2][attr3][val1][val2].append([])
                            
    for attr1 in range(NumAttr-1):
        essences2.append([])
        for attr2 in range(NumAttr-1-attr1):
            essences2[attr1].append([])
            for val1 in range(NumEVals):
                essences2[attr1][attr2].append([])
                for val2 in range(NumEVals):
                    essences2[attr1][attr2][val1].append([])


class Requirement:
    def __init__ (self, mag, sign1, attr1, sign2, attr2, sign3=None, attr3=None):
        self.mag = mag
        self.prefix = "| "
        self.signs = []
        self.attrs = []

        self.attrs.append(attr1)
        self.prefix += AttrShortNames[attr1]
        if(sign1 == 1):
            self.signs.append(1)
            self.prefix += "+"*(mag+2)
        else:
            self.signs.append(-1)
            self.prefix += "-"*(mag+2)

        self.attrs.append(attr2)
        self.prefix += " || " + AttrShortNames[attr2+attr1+1]
        if(sign2 == 1):
            self.signs.append(1)
            self.prefix += "+"*(mag+2)
        else:
            self.signs.append(-1)
            self.prefix += "-"*(mag+2)
        
        if(sign3 != None and attr3 != None):
            self.attrs.append(attr3)
            self.prefix += " || " + AttrShortNames[attr3+attr2+attr1+2]
            if(sign3 == 1):
                self.signs.append(1)
                self.prefix += "+"*(mag+2)
            else:
                self.prefix += "-"*(mag+2)
                self.signs.append(-1)
        
        self.recipes = []
        #self.count = 0

    def findRecipe(self, essences2, essences3):
        cheapest = []
        cheapVal = 10000000
        target = self.mag*3 + 3
        first = True
        
        if len(self.signs) == 3:
            nAttr=3
            cur = [[]]*nAttr
            
            cur[0] = [3*self.signs[0]]*NumE
            cur[0][NumE-1] += self.signs[0]
            while(self.decrementSortedList(cur[0], target, self.signs[0])):
                cur[1] = [3*self.signs[1]]*NumE
                cur[1][NumE-1] += self.signs[1]
                while(self.decrementSortedList(cur[1], target, self.signs[1])):
                    cur[2] = [3*self.signs[2]]*NumE
                    cur[2][NumE-1] += self.signs[2]
                    #print self.count
                    
                    while(self.decrementSortedList(cur[2], target, self.signs[2])):
                        # grab 5 essences from essences3[self.attrs[0]][self.attrs[1]][self.attrs[2]][cur[0][i]][cur[1][i]][cur[2][i]]
                        l = cur[0]
                        m = cur[1][:]
                        n = cur[2][:]
                        done = False
                        while not done:
                            permute_next(m)
                            if(m == cur[1]):
                                done = True
                            done2 = False
                            while not done2:
                                permute_next(n)
                                if n == cur[2]:
                                    done2 = True
                                #if l == [-3, -2, -2, 0, 0] and m == [-1, -3, 0, -2, -3] and n == [-2, -1, -3, -1, -2]:
                                    #print l, m, n
                                found = False
                                #self.count += 1
                                possRec = []
                                sum = 0
                                for i in range(NumE):
                                    found = False
                                    #if l == [-3, -2, -2, 0, 0] and m == [-1, -3, 0, -2, -3] and n == [-2, -1, -3, -1, -2]:
                                        #print i
                                    for e in essences3[self.attrs[0]][self.attrs[1]][self.attrs[2]][l[i]+3][m[i]+3][n[i]+3]:
                                        #if l == [-3, -2, -2, 0, 0] and m == [-1, -3, 0, -2, -3] and n == [-2, -1, -3, -1, -2]:
                                            #print e
                                        if not e in possRec:
                                            found = True
                                            possRec.append(e)
                                            sum += e.cost
                                            break
                                    if not found:
                                        break
                                if found and sum < cheapVal:
                                    cheapVal = sum
                                    cheapest = possRec
            self.recipes.append(cheapest)
        else:
            nAttr=2
            cur = [[]]*nAttr
            
            cur[0] = [3*self.signs[0]]*NumE
            cur[0][NumE-1] += self.signs[0]
            while(self.decrementSortedList(cur[0], target, self.signs[0])):
                cur[1] = [3*self.signs[1]]*NumE
                cur[1][NumE-1] += self.signs[1]
                while(self.decrementSortedList(cur[1], target, self.signs[1])):
                    # grab 5 essences from essences3[self.attrs[0]][self.attrs[1]][self.attrs[2]][cur[0][i]][cur[1][i]][cur[2][i]]
                    l = cur[0]
                    m = cur[1][:]
                    done = False
                    while not done:
                        permute_next(m)
                        if(m == cur[1]):
                            done = True
                        found = False
                        #self.count += 1
                        possRec = []
                        sum = 0
                        for i in range(NumE):
                            found = False
                            for e in essences2[self.attrs[0]][self.attrs[1]][l[i]+3][m[i]+3]:
                                if not e in possRec:
                                    found = True
                                    possRec.append(e)
                                    sum += e.cost
                                    break
                            if not found:
                                break
                        if found and sum < cheapVal:
                            cheapVal = sum
                            cheapest = possRec
            self.recipes.append(cheapest)
                            
    def decrementSortedList(self, list, target, sign):
        list[4] -= sign
        if list[4]*-1*sign > 3 or not self.checkValues(list, target, sign):
            list[3] -= sign
            list[4] = list[3]
            if list[3]*-1*sign > 3 or not self.checkValues(list, target, sign):
                list[2] -= sign
                list[3] = list[2]
                list[4] = list[2]
                if list[2]*-1*sign > 3 or not self.checkValues(list, target, sign):
                    list[1] -= sign
                    list[2] = list[1]
                    list[3] = list[1]
                    list[4] = list[1]
                    if list[1]*-1*sign > 3 or not self.checkValues(list, target, sign):
                        list[0] -= sign
                        list[1] = list[0]
                        list[2] = list[0]
                        list[3] = list[0]
                        list[4] = list[0]
                        if list[0]*-1*sign > 3 or not self.checkValues(list, target, sign):
                            return False
        return True
    
    def decrementUnsortedList(self, list, target, sign):
        list[4] -= sign
        if list[4]*-1*sign > 3 or not self.checkValues(list, target, sign):
            list[3] -= sign
            list[4] = 3*sign
            if list[3]*-1*sign > 3 or not self.checkValues(list, target, sign):
                list[2] -= sign
                list[3] = 3*sign
                list[4] = 3*sign
                if list[2]*-1*sign > 3 or not self.checkValues(list, target, sign):
                    list[1] -= sign
                    list[2] = 3*sign
                    list[3] = 3*sign
                    list[4] = 3*sign
                    if list[1]*-1*sign > 3 or not self.checkValues(list, target, sign):
                        list[0] -= sign
                        list[1] = 3*sign
                        list[2] = 3*sign
                        list[3] = 3*sign
                        list[4] = 3*sign
                        if list[0]*-1*sign > 3 or not self.checkValues(list, target, sign):
                            return False
        return True


    def checkValues(self, list, target, sign):
        sum = 0
        for j in range(NumE):
            sum += list[j]
        if sum*sign > target:
            return True
        else:
            return False


    def __str__ (self):
        result = ""
        if len(self.recipes) == 0:
            result += self.prefix
            result += "\n|-"
        else:
            first = True
            for r in self.recipes:
                if not first:
                    result += "\n"
                else:
                    first = False
                result += self.prefix
                for e in r:
                    result += " || " + str(e)
                result += "\n|-"
        return result

    def __cmp__ (self, other):
        assert isinstance(other, Requirement)
        result = cmp(self.mag, other.mag)
        if result == 0:
            result = cmp(len(self.attrs), len(other.attrs))
            if result == 0:
                result = cmp(self.attrs[0], other.attrs[0])
                if result == 0:
                    result = cmp(other.signs[0], self.signs[0])
                    if result == 0:
                        result = cmp(self.attrs[1], other.attrs[1])
                        if result == 0:
                            result = cmp(other.signs[1], self.signs[1])
                            if result == 0 and len(self.attrs)==3:
                                result = cmp(self.attrs[2], other.attrs[2])
                                if result == 0:
                                    result = cmp(other.signs[2], self.signs[2])
        return result

# Populate all possible requirements
def setRequirements(requirements2, requirements3):
    for mag in range(1,3):
        for sign1 in range(2):
            for sign2 in range(2):
                for sign3 in range(2):
                    for attr1 in range(NumAttr-2):
                        for attr2 in range(NumAttr-2-attr1):
                            for attr3 in range(NumAttr-2-attr2-attr1):
                                requirements3.append(Requirement(mag, sign1, attr1, sign2, attr2, sign3, attr3))

    for mag in range(4):
        for sign1 in range(2):
            for sign2 in range(2):
                for attr1 in range(NumAttr-1):
                    for attr2 in range(NumAttr-1-attr1):
                        requirements2.append(Requirement(mag, sign1, attr1, sign2, attr2))

    requirements = requirements2 + requirements3
    requirements.sort()
    return requirements


def ReadFromFile(filename, essences2, essences3):
    f = open(filename, "r")
    line = f.readline()
    while(line != ''):
        match = essence_line_pattern.match(line)
        if match != None:
            e = Essence(match, essences2, essences3)
        else:
            print "Did not match: \"" + line,
        line = f.readline()

# Sorts lists of Essences inside a table cell by cost
def sortEssences(essences2, essences3):
    for attr1 in range(NumAttr-2):
        for attr2 in range(NumAttr-2-attr1):
            for attr3 in range(NumAttr-2-attr2-attr1):
                for val1 in range(NumEVals):
                    for val2 in range(NumEVals):
                        for val3 in range(NumEVals):
                            essences3[attr1][attr2][attr3][val1][val2][val3].sort()

    for attr1 in range(NumAttr-1):
        for attr2 in range(NumAttr-1-attr1):
            for val1 in range(NumEVals):
                for val2 in range(NumEVals):
                    essences2[attr1][attr2][val1][val2].sort()

def printEssenceValueTable(table):
    if len(table[0][0]) == 7:
        for attr1 in range(NumAttr-1):
            firstName = AttrNames[attr1]
            for attr2 in range(NumAttr-1-attr1):
                secondName = AttrNames[attr2+attr1+1]
                for val1 in range(NumEVals):
                    for val2 in range(NumEVals):
                        print val1-3, firstName, val2-3, secondName + ":"
                        for e in table[attr1][attr2][val1][val2]:
                            print e
    else:
        for attr1 in range(NumAttr-2):
            firstName = AttrNames[attr1]
            for attr2 in range(NumAttr-2-attr1):
                secondName = AttrNames[attr2+attr1+1]
                for attr3 in range(NumAttr-2-attr2-attr1):
                    thirdName = AttrNames[attr3+attr2+attr1+2]
                    for val1 in range(NumEVals):
                        for val2 in range(NumEVals):
                            for val3 in range(NumEVals):
                                print val1-3, firstName, val2-3, secondName, val3-3, thirdName + ":"
                                for e in table[attr1][attr2][attr3][val1][val2][val3]:
                                    print e

def findRecipesFromQueue(inq, outq, essences2, essences3):
    while not inq.empty():
        try:
            r = inq.get(True, 2)
            r.findRecipe(essences2, essences3)
            print r
            outq.put(r)
            inq.task_done()
        except Empty:
            pass
    print "done"
    #outq.close()

def startProcs(procs, inq, outq,essences2,essences3):
    numProc = 4
    for i in range(numProc):
        p = Process(target=findRecipesFromQueue, args=(inq,outq,essences2,essences3))
        procs.append(p)
        p.start()

def processAllReqs(requirements, inq, outq, essences2, essences3):
    for r in requirements:
        # bellow is "if not(SET)"
        if not (len(r.attrs) == 3 and r.mag == 1):
            inq.put(r)
    #inq.close()
    procs=[]
    startProcs(procs, inq, outq, essences2, essences3)
    inq.join()
    tmp = []
    while not outq.empty():
        tmp.append(outq.get())
        outq.task_done()
    cont = True
    for p in procs:
        print "Joining", p.pid
        p.join()
    #while cont:
    #    cont = False
    #    for p in procs:
    #        if p.is_alive() == True:
    #            cont = True
    #    print ".",
    #    time.sleep(1)
    tmp.sort()
    print "\n***\n"
    for r in tmp:
        print r

def quickTest(requirements, inq, outq, essences2, essences3):
    inq.put(requirements[0])
    inq.put(requirements[1])
    inq.put(requirements[2])
    inq.put(requirements[3])
    startProcs(inq, outq, essences2, essences3)
    inq.join()
    tmp = []
    while not outq.empty():
        tmp.append(outq.get())
    tmp.sort()
    print "\n***\n"
    for r in tmp:
        print r

def pointTest(essences):
    l = [-3, -2, -2, 0, 0]
    m = [-1, -3, 0, -2, -3]
    n = [-2, -1, -3, -1, -2]
    
    for i in range(5):
        for e in essences[0][0][0][l[i]+3][m[i]+3][n[i]+3]:
            print i,e

def m():
    essences2 = []
    essences3 = []

    Requirements2 = []
    Requirements3 = []
    
    inQue = JoinableQueue()
    outQue = JoinableQueue()

    buildEssenceTables(essences2, essences3)
    Requirements = setRequirements(Requirements2, Requirements3)
    ReadFromFile("essences.txt",essences2, essences3)
    sortEssences(essences2, essences3)

    #printEssenceValueTable(essences3)
    #printEssenceValueTable(essences2)
    #quickTest(Requirements3, inQue, outQue, essences2, essences3)
    #pointTest()
    processAllReqs(Requirements, inQue, outQue, essences2, essences3)


def f(name):
    print 'hello', name

if __name__ == '__main__':
    m()
    #p = Process(target=f, args=('bob',))
    #p.start()
    #p.join()
    #print "bye"
